/**
 * Copyright (c) Members of the EGEE Collaboration. 2007-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2007-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl> 
 *
 *
 */


/* network functions */
/*
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <unistd.h>
*/

#include <errno.h>
#include "net_common.h"
#include "scas_log.h"



/****************************************************
Function: getIPString
Description:
          Function returning the IP string representation of a socket address

          This function originated from the EGEE-III SA2 team as i
          IPv6 example and is altered to work of its new context
Return:
    void
****************************************************/
void getIPString(char *ip, 
                 int ip_length, 
                 struct sockaddr *paddress)
{
    /* retrieve the ip */
    getnameinfo(paddress, sizeof(struct sockaddr_storage), ip, ip_length, 0, 0, NI_NUMERICHOST);

    /* if ipv4-mapped address, we remove the mapping notation in order to get only  */
    /* the IPv4 address */
    if (strncasecmp(ip, PREFIX_V4_MAPPED, strlen(PREFIX_V4_MAPPED)) == 0)
    {
        memmove(ip, 
                ip + strlen(PREFIX_V4_MAPPED),
                strlen(ip) - strlen(PREFIX_V4_MAPPED) + 1);
    }
}


/****************************************************
Function: net_addr2host
Description:
          Will convert a sockaddr_storage struct to
          a hostname, through reverse Host and DNS lookup
          or it will return the IP address as string.

          This function originated from the EGEE-III SA2 team as i
          IPv6 example and is altered to work of its new context
Variables:
            char ** hostname            : buffer in which the hostname will be written,
                                          Will allocate if not allocated yet.
            int     bufsize             : size of hostname buffer
            struct  sockaddr_storage ss : socket address information (for Internet)
Return:
            0 : Good
            !0 : Not good
****************************************************/
int net_addr2host (char **                    hostname,
                   int                        bufsize,
                   struct sockaddr_storage *  ss)
{
    struct hostent * hp = NULL;
    int rc = 0;

    /* Allocate if you didn't yet */
    if (*hostname == NULL)
    {
        *hostname = (char *) malloc (sizeof (char) * bufsize);
    }


    /* Try to get a hostname resolution from the IPv6 compliant sockaddr container */
    if ((rc = getnameinfo (ss, sizeof (struct sockaddr_storage),
                           *hostname, bufsize, 
                           NULL, 0, 
                           NI_NAMEREQD) != 0))
    {
        switch (rc)
        {
            case EAI_AGAIN :
                scas_log (5, "%s: The specified name cannot be currently resolved. Further attempts might succeed.\n", __func__);
                break;
            case EAI_BADFLAGS :
                scas_log (5, "%s: The flags parameter is set to an invalid value.\n", __func__);
                break;
            case EAI_FAIL :
                scas_log (5, "%s: A fatal error (with no recovery possibility) occurred.\n", __func__);
                break;
            case EAI_FAMILY :
                scas_log (5, "%s: The specified socket address structure contained an unrecognized address family or the address length was not valid for the specified family.\n", __func__);
                break;
            case EAI_MEMORY :
                scas_log (5, "%s: A memory allocation failure occurred while attempting to set aside space for a return value.\n", __func__);
                break;
            case EAI_NONAME :
                scas_log (5, "%s: NI_NAMEREQD is set and getnameinfo() cannot find the host's name.\n", __func__);
                break;
            case EAI_OVERFLOW :
                scas_log (5, "%s: The buffer pointed to by the node parameter or the service parameter was too small and overflowed.\n", __func__);
                break;
            case EAI_SYSTEM :
                scas_log (5, "%s: A system error occurred; the errno variable has been set to indicate that error.\n", __func__);
                break;
            default :
                scas_log (5, "%s: Some other problem... rc = %d\n", __func__, rc);
                break;
        }


        /* Map the IP address to a host name. */
        hp = gethostbyaddr (ss, sizeof(struct sockaddr_storage), ss->ss_family);
        if (hp)
        {
            strncpy(*hostname, hp->h_name, bufsize);
            return 0;
        }
        else
        {
            /* If an address resolution couldn't result into a name, then we can use the IP Address */
            scas_log (1, "DEBUG: couldn't get a hostname\n");
            getIPString(*hostname, bufsize, ss);
        }
    }
    return 0;
}



/****************************************************
Function: addr2host
Variables:
          char ** hostname         : buffer in which the hostname will be written (requires preallocation)
          int     bufsize          : size of hostname buffer
          struct sockaddr_in  addr : socket address information (for Internet)
Return:
    0 : Good
   !0 : Not good
****************************************************/
#ifdef DEPRECATED_CODE
int addr2host (char **            hostname,
               int                bufsize,
               struct sockaddr_in addr)
{
    struct hostent * hp = NULL;

    /* Map the IP address to a host name. */
    hp = gethostbyaddr ((char *)&addr.sin_addr, sizeof(struct in_addr), addr.sin_family);
    if (hp)
        strncpy(*hostname, hp->h_name, bufsize);
    else
        strncpy(*hostname, inet_ntoa(addr.sin_addr), bufsize);

    return 0;
}
#endif /* DEPRECATED_CODE */





int create_client_socket (int * client_socket, 
                          const char * server, 
                          int port, 
                          int time_out_milliseconds)
{
    char                               *logstr = "create_client_socket";
    struct addrinfo                     hints;
    struct addrinfo                    *res;
    int                                 rc;
    int                                 mysock = -1;
    char                                portstr[24];

    struct timeval *                    wait_tv = NULL;
    struct timeval                      preset_tv;
    unsigned int                        preset_tvlen = sizeof preset_tv;


    memset(&hints, 0, sizeof(struct addrinfo));

    hints.ai_socktype = SOCK_STREAM;
    hints.ai_family   = AF_INET;

    /* Get addrinfo */
    snprintf(portstr, 24, "%d", port);
    rc = getaddrinfo(server, &portstr[0], &hints, &res);
    if (rc != 0)
    {
        scas_log (1, "Error: Failed to getaddrinfo (%s, %s, *, *)\n", server, portstr);
        return 1;
    }


    /* Create new socket */
    if ((mysock = socket(res->ai_family, res->ai_socktype, res->ai_protocol)) < 0)
    {
        scas_log (1, "Error: Failed to create socket\n");
        return 1;
    }

    
    /* Grab timeout setting */
    if (getsockopt (mysock, SOL_SOCKET, SO_RCVTIMEO, (char *)&preset_tv, &preset_tvlen) < 0)
    {
        scas_log (1, "%s: Error: Failed to get the timeout setting\n", logstr);
        return 1;
    }


    /* Set connection timeout on the socket */
    wait_tv = (struct timeval *) malloc (sizeof (struct timeval));
    wait_tv->tv_sec = (time_out_milliseconds - (time_out_milliseconds % 1000)) / 1000;
    wait_tv->tv_usec = (time_out_milliseconds % 1000) * 1000;
    if (setsockopt (mysock, SOL_SOCKET, SO_RCVTIMEO, (char *)wait_tv, sizeof *wait_tv) < 0)
    {
        scas_log (1, "%s: Error: Failed to set the timeout setting\n", logstr);
        return 1;
    }
    free (wait_tv);
    wait_tv = NULL;


    /* Connecting socket to host on port with timeout */
    if (connect(mysock, res -> ai_addr, res -> ai_addrlen) < 0)
    {
        scas_log_debug (1, "Failed to connect\n");
        return 1;
    }
    else
    {
        /* Socket is succesfuly connected */
        setsockopt (mysock, SOL_SOCKET, SO_KEEPALIVE, 0, 0);

        *client_socket = mysock;
        return 0;
    }

    /* Failure */
    return 1;
}

