/**
 * Copyright (c) Members of the EGEE Collaboration. 2007-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2007-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl> 
 *
 *
 */

#include "pdp_xacml_lcas_lcmaps.h"
#include "scas_log.h"

/* impl */



/**********************************************
Function: 
             execLCAS  
Description:
             Executing LCAS, getting authorizational decision in return 
**********************************************/
int execLCAS (xacml_profile_request_t * xacml_profile_request)
{
    #define GSS_C_NO_CREDENTIAL ((gss_cred_id_t) 0)

    int          rc        = 0;
    char *       rsl       = NULL;
    lcas_request_t request = NULL;
    char *       user_dn   = NULL;
    const char * logstr    = "execLCAS";

    char *       lcas_log_file_env = NULL;

    if (!xacml_profile_request)
    {
        scas_log (1, "%s: No XACML Request to process in LCAS. Can't continue with this request.\n", logstr);
        return 100;
    }

    user_dn     = xacml_profile_request -> xacml_profile_subject.xacml_subject_x509_id.subject_x509_id;
    rsl         = xacml_profile_request -> xacml_profile_action.xacml_action_rsl_string.action_rsl_string;


    /* Initialize */
    /* When set, use LCAS_LOG_FILE */
    if ( (lcas_log_file_env = getenv("LCAS_LOG_FILE"))!=NULL)	{
        if (lcas_init_and_logfile(lcas_log_file_env, NULL, DO_USRLOG))
        {
            scas_log (1, "%s: Initialization of LCAS failed!\n", logstr);
            return 101;
        }
	else
	    scas_log_debug (5, "LCAS logs to %s\n", lcas_log_file_env);
    }
    else /* LCAS_LOG_FILE is unset */
    {
	if (lcas_init_and_logfile(NULL, scas_get_log_file_pointer(), scas_get_log_type())!=0)	{
            scas_log (1, "%s: Initialization of LCAS failed!\n", logstr);
            return 101;
        } else
	    scas_log_debug (5, "LCAS logs successfully integrated with SCAS log file\n");
    }


    /* LCAS run */
    request = (lcas_request_t) rsl;
    if (lcas_get_fabric_authorization (user_dn, GSS_C_NO_CREDENTIAL, request))
    {
        scas_log (1, "%s: LCAS failed authorization.\n", logstr);
        rc = 102;
    }

    /* LCAS term */
    if (lcas_term())
    {
        scas_log (1, "%s: LCAS termination failure.\n", logstr);
        rc = 103;
    }

    return rc;
}


/**********************************************
Function: 
             execLCMAPS
Description:
             Executing LCMAPS, getting a mapping in return 
**********************************************/
int execLCMAPS (xacml_profile_request_t * xacml_profile_request, lcmaps_account_info_t * paccount_info)
{
    int rc = 0; 
    int i = 0; 

    char *            user_dn     = NULL;
    char **           fqan_list   = NULL;
    int               nfqan       = 0;
    lcmaps_request_t  request     = NULL;
    int               npols       = 0;
    char **           policynames = NULL;
    uid_t             uid         = -1;
    gid_t *           pgid_list   = NULL;
    int               npgid       = 0;
    gid_t *           sgid_list   = NULL;
    int               nsgid       = 0;
    char *            poolindex   = NULL;

    char *            lcmaps_log_file_env = NULL;


    if (!xacml_profile_request)
    { 
        scas_log (1, "execLCMAPS: No XACML Request to process in LCMAPS. Can't continue with this request.\n");
        return 100;
    } 


    user_dn     = xacml_profile_request -> xacml_profile_subject.xacml_subject_x509_id.subject_x509_id;
    nfqan       = xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan_cnt;
    /* fqan_list   = malloc (sizeof (char *) * nfqan); */
    fqan_list   = calloc (nfqan, sizeof (char *));
    for (i = 0; i < nfqan; i++)
        fqan_list[i] = strdup(xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan[i].subject_voms_fqan);


    /* if (lcmaps_init_and_logfile(getenv("LCMAPS_LOG_FILE"), NULL, DO_SYSLOG)) */
    /* if (lcmaps_init_and_logfile(getenv("LCMAPS_LOG_FILE"), NULL, DO_USRLOG)) */

    /* When set, use LCMAPS_LOG_FILE */
    if ( (lcmaps_log_file_env = getenv("LCMAPS_LOG_FILE"))!=NULL) {
	if ( lcmaps_init_and_logfile(lcmaps_log_file_env, NULL, DO_USRLOG)!=0 )
	{
            scas_log (1, "execLCMAPS: Initialization of LCMAPS failed!\n");
            return 101;
	}
	else
	    scas_log_debug (5, "LCMAPS logs to %s\n", lcmaps_log_file_env);
    }
    else /* LCMAPS_LOG_FILE is unset */
    {
	if (lcmaps_init_and_logfile(NULL, scas_get_log_file_pointer(), scas_get_log_type())!=0 )    {
	    scas_log (1, "execLCMAPS: Initialization of LCMAPS failed!\n");
            return 101;
	} else
	    scas_log_debug (5, "LCMAPS logs successfully integrated with SCAS log file\n");
    }

    if ((rc = lcmaps_account_info_init(paccount_info)) != 0) 
    {
        scas_log (1, "execLCMAPS: Error initializing LCMAPS\n");
        return 102;
    }


    if ((rc = lcmaps_run_with_fqans_and_return_account (user_dn, 
                                                        fqan_list, nfqan, 
                                                        request, 
                                                        npols, policynames,
                                                        &uid, 
                                                        &pgid_list, &npgid, 
                                                        &sgid_list, &nsgid, 
                                                        &poolindex)) != 0)
    {
        scas_log (1, "execLCMAPS: LCMAPS failed to return a mapping result\n");
        rc = 103;
    } 
    else
    {
        scas_log (5, "execLCMAPS: LCMAPS resulted into a succesfull mapping\n");

        /* Getting the LCMAPS information back into the thread */
        paccount_info->uid       = uid;
        paccount_info->npgid     = npgid; 
        paccount_info->nsgid     = nsgid; 
        if (poolindex)
            paccount_info->poolindex = strdup(poolindex);
        /* paccount_info->pgid_list = malloc (sizeof (gid_t) * npgid); */
        /* paccount_info->sgid_list = malloc (sizeof (gid_t) * nsgid); */
        paccount_info->pgid_list = calloc (npgid, sizeof (gid_t));
        paccount_info->sgid_list = calloc (nsgid, sizeof (gid_t));
        for (i = 0; i < npgid; i++) paccount_info->pgid_list[i] = pgid_list[i];
        for (i = 0; i < nsgid; i++) paccount_info->sgid_list[i] = sgid_list[i];
    }

    /* Terminate LCMAPS */
    lcmaps_term();

    /* Liberate memory */
    for (i = 0; i < nfqan; i++)
        free(fqan_list[i]);
    if (fqan_list)
        free(fqan_list);

    return rc;
}


