/**
 * Copyright (c) Members of the EGEE Collaboration. 2007-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2007-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl> 
 *
 *
 */


/*!
    \file   scas_log.c
    \brief  Logging routines for SCAS
    \author Martijn Steenbakkers for the EU DataGrid.
*/


/*****************************************************************************
                            Include header files
******************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include <syslog.h>
#include <time.h>
#include <ctype.h>
#include <unistd.h>


#include "_scas_log.h"


/******************************************************************************
                          Module specific prototypes
******************************************************************************/
#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL 0 /*!< default debugging level */
#endif /* DEBUG_LEVEL */

/******************************************************************************
                       Define module specific variables
******************************************************************************/
static FILE *  scas_logfp=NULL; /*!< stream associated with logfile \internal */
static int     logging_usrlog=0; /*!< flag to do user logging \internal */
static int     logging_syslog=0; /*!< flag to use syslog \internal */
//static int     debug_level=DEBUG_LEVEL; /*!< debugging level \internal */
static int     debug_level=0; /*!< debugging level \internal */
static char *  extra_logstr = NULL; /*!< string to be included in every log statement \internal */ 
static int     should_close_scas_logfp = 0; /*!< Flag to check if the log stream should be closed \internal */ 
static char *  scas_log_path = NULL;  /*!< String that holds the path to the logfile on the file system \internal */

/******************************************************************************
Function:       scas_log_open()
Description:    Start logging
Parameters:
                path:    path of logfile
                fp:      file pointer to already opened file (or NULL)
                logtype: DO_USRLOG, DO_SYSLOG
Returns:        0 succes
                1 failure
******************************************************************************/
/*!
    \fn scas_log_open(
        char * path,
        FILE * fp,
        unsigned short logtype
        )
    \brief Start logging

    This function should only be used by the SCAS itself.
    It opens the logfile and tries to set the debugging level in the following order:
    -# Try if DEBUG_LEVEL > 0
    -# Try if environment variable SCAS_DEBUG_LEVEL is set and if it is an integer > 0
    -# Otherwise set debug_level = 0;

    \param path    path of logfile.
    \param fp      file pointer to already opened file (or NULL)
    \param logtype DO_USRLOG, DO_SYSLOG

    \retval 0 succes.
    \retval 1 failure.
    \internal
*/
int
scas_log_open(char * path, FILE * fp, unsigned short logtype )
{
    char * debug_env = NULL;
/*    char * logstr_env = NULL;*/

    if ((logtype & DO_SYSLOG) == DO_SYSLOG)
    {
//        fprintf(stderr,"scas_log_open() error: attempt to do syslogging,");
//        fprintf(stderr," not supported yet\n");
        logging_syslog=1;
#if 0
        /* Not yet applicable, wait for SCAS to become daemon */
        openlog("EDG SCAS", LOG_PID, LOG_AUTHPRIV);
#endif
    }
    if ((logtype & DO_USRLOG) == DO_USRLOG)
    {
        logging_usrlog=1;
        if (fp != NULL)
        {
            /* File already opened, should not be closed at the end */
            scas_logfp=fp;
            should_close_scas_logfp = 0;
        }
        else if (path != NULL)
        {
            /* Try to append to file */
            if ((scas_logfp = fopen(path, "a")) == NULL)
            {
                fprintf(stderr, "scas_log_open(): Cannot open logfile %s: %s\n",
                        path, strerror(errno));
                if (logging_syslog)
                {
                    syslog(LOG_ERR, "scas_log_open(): Cannot open logfile %s\n", path);
                }
                return 1;
            }
            should_close_scas_logfp = 1;
            scas_log_path = path;
        }
        else
        {
            fprintf(stderr, "scas_log_open(): Please specify either (open) file descriptor");
            fprintf(stderr, " or name of logfile\n");
            return 1;
        }
    }
    /*
     * Set the debugging level:
     *    1. Try if DEBUG_LEVEL > 0
     *    2. Try if SCAS_DEBUG_LEVEL is set and if it is an integer
     *    3. set debug_level = 0;
     */
    if ( (int)(DEBUG_LEVEL) > 0 )
    {
        debug_level = (int)(DEBUG_LEVEL);
    }
    else if ( (debug_env = getenv("SCAS_DEBUG_LEVEL")) )
    {
        /* convert into integer */
        int j = 0;

        for (j = 0; j < strlen(debug_env); j++)
        {
            if (!isdigit((debug_env)[j]))
            {
                fprintf(stderr,"scas_log_open(): found non-digit in environment variable in \"SCAS_DEBUG_LEVEL\" = %s\n", debug_env);
                return 1;
            }
        }
        debug_level = atoi(debug_env);
        if (debug_level < 0)
        {
            fprintf(stderr,"scas_log_open(): environment variable in \"SCAS_DEBUG_LEVEL\" should be >= 0\n");
            return 1;
        }
    }
    else
    {
        debug_level = 0;
    }

    if (debug_level > 0)
    {
        scas_log (3,"scas_log_open(): setting debugging level to %d\n", debug_level);
    }

    /* Don't look at the external string value to prefix the log messages. */
    return 0;
}

/******************************************************************************
Function:       scas_log()
Description:    Log information to file and or syslog
Parameters:
                prty:    syslog priority (if 0 don't syslog)
                fmt:     string format
Returns:        0 succes
                1 failure
******************************************************************************/
/*!
    \fn scas_log(
        int prty,
        char * fmt,
        ...
        )
    \brief log information
    
    This function logs information for SCAS and its plugins.
    Syslog() is called with the specified priority. No syslog() is done if the
    priority is 0.

    \param prty syslog priority (if 0 don't syslog).
    \param fmt  string format
    \param ...  variable argument list

    \retval 0 succes.
    \retval 1 failure.
*/
int
scas_log(int prty, char * fmt, ...)
{
    va_list pvar;
    char    buf[MAX_LOG_BUFFER_SIZE];
    int     res;

    va_start(pvar, fmt);
    res=vsnprintf(buf,MAX_LOG_BUFFER_SIZE,fmt,pvar);
    va_end(pvar);
    if ( (res >= MAX_LOG_BUFFER_SIZE) || (res < 0) )
    {
        fprintf(stderr,"scas_log(): log string too long (> %d)\n",
                MAX_LOG_BUFFER_SIZE);
    }
    if (logging_usrlog)
    {
        if (scas_logfp == NULL)
        {
            fprintf(stderr,"scas_log() error: cannot open file descriptor\n");
            return 1;
        }
        if (extra_logstr == NULL)
        {
            fprintf(scas_logfp,"SCAS %d[%d]: %s", prty, getpid(), buf);
        }
        else
        {
            fprintf(scas_logfp,"SCAS %d[%d]: %s : %s", prty, getpid(), extra_logstr, buf);
        }
        fflush(scas_logfp);
    }
    if (logging_syslog && prty)
    {
        syslog(prty, buf);
    }

    return 0;
}

/******************************************************************************
Function:       scas_log_a_string()
Description:    Log a string according to the passed format to file and or syslog
Parameters:
                prty:       syslog priority (if 0 don't syslog)
                fmt:        string format
                the_string: the string
Returns:        0 succes
                1 failure
******************************************************************************/
/*!
    \fn scas_log_a_string(
        int prty,
        char * fmt,
        char * the_string
        )
    \brief log information
    
    This function logs information for SCAS and its plugins.
    Syslog() is called with the specified priority. No syslog() is done if the
    priority is 0.

    \param prty         syslog priority (if 0 don't syslog).
    \param fmt          string format
    \param the_string   the string

    \retval 0 succes.
    \retval 1 failure.
*/
int
scas_log_a_string(int prty, char * fmt, char * the_string)
{
    char    buf[MAX_LOG_BUFFER_SIZE];
    int     res;

    res = snprintf(buf, MAX_LOG_BUFFER_SIZE, fmt, the_string);
    if ( (res >= MAX_LOG_BUFFER_SIZE) || (res < 0) )
    {
        fprintf(stderr,"scas_log_a_string(): log string too long (> %d)\n",
                MAX_LOG_BUFFER_SIZE);
    }
    if (logging_usrlog)
    {
        if (scas_logfp == NULL)
        {
            fprintf(stderr,"scas_log() error: cannot open file descriptor\n");
            return 1;
        }
        if (extra_logstr == NULL)
        {
            fprintf(scas_logfp,"SCAS %d: %s", prty, buf);
        }
        else
        {
            fprintf(scas_logfp,"SCAS %d: %s : %s", prty, extra_logstr, buf);
        }
        fflush(scas_logfp);
    }
    if (logging_syslog && prty)
    {
        syslog(prty, buf);
    }

    return 0;
}

/******************************************************************************
Function:       scas_log_debug()
Description:    Print debugging information
Parameters:
                debug_lvl: debugging level
                fmt:       string format
Returns:        0 succes
                1 failure
******************************************************************************/
/*!
    \fn scas_log_debug(
        int debug_lvl,
        char * fmt,
        ...
        )
    \brief Print debugging information

    This function prints debugging information (using scas_log with priority 0)
    provided debug_lvl <= DEBUG_LEVEL (default is 0).

    \param debug_lvl debugging level
    \param fmt       string format
    \param ...       variable argument list

    \retval 0 succes.
    \retval 1 failure.
*/
int
scas_log_debug(int debug_lvl, char * fmt, ...)
{
    va_list pvar;
    char    buf[MAX_LOG_BUFFER_SIZE];
    int     res;

    va_start(pvar, fmt);
    res=vsnprintf(buf,MAX_LOG_BUFFER_SIZE,fmt,pvar);
    va_end(pvar);
    if ( (res >= MAX_LOG_BUFFER_SIZE) || (res < 0) )
    {
        fprintf(stderr,"scas_log(): log string too long (> %d)\n",
                MAX_LOG_BUFFER_SIZE);
    }
    if (debug_lvl <= debug_level)
    {
        scas_log(0,buf);
        return 0;
    }
    return 1;
}

/******************************************************************************
Function:       scas_log_a_string_debug()
Description:    Print debugging information
Parameters:
                debug_lvl:  debugging level
                fmt:        string format
                the_string: the string
Returns:        0 succes
                1 failure
******************************************************************************/
/*!
    \fn scas_log_a_string_debug(
        int debug_lvl,
        char * fmt,
        char * the_string
        )
    \brief Print debugging information

    This function prints debugging information (using scas_log with priority 0)
    provided debug_lvl <= DEBUG_LEVEL (default is 0).

    \param debug_lvl    debugging level
    \param fmt          string format
    \param the_string   the string

    \retval 0 succes.
    \retval 1 failure.
*/
int
scas_log_a_string_debug(int debug_lvl, char * fmt, char * the_string)
{
    if (debug_lvl <= debug_level)
    {
        scas_log_a_string(0, fmt, the_string);
        return 0;
    }
    return 1;
}

/******************************************************************************
Function:       scas_log_close()
Description:    Stop logging
Parameters:
Returns:        0 succes
                1 failure
******************************************************************************/
/*!
    \fn scas_log_close()
    \brief Stop logging.

    \internal
    This function should only be used by the SCAS itself.

    \retval 0 succes.
    \retval 1 failure.
    \internal
*/
int
scas_log_close()
{
    if (extra_logstr != NULL)
    {
        free(extra_logstr);
        extra_logstr = NULL;
    }

    if (should_close_scas_logfp)
    {
        fclose(scas_logfp);
        scas_logfp=NULL;
    }

    return 0;
}


/******************************************************************************
Function:       scas_log_time()
Description:    Log information to file and or syslog with a timestamp
Parameters:
                prty:    syslog priority (if 0 don't syslog)
                fmt:     string format
Returns:        0 succes
                1 failure
******************************************************************************/
/*!
    \fn scas_log_time(
        int prty,
        char * fmt,
        ...
        )
    \brief log information with timestamp
 
    This function logs information with a timestamp for SCAS and its plugins.
    Syslog() is called with the specified priority. No syslog() is done if the
    priority is 0.
 
    \param prty syslog priority (if 0 don't syslog).
    \param fmt  string format
    \param ...  variable argument list
 
    \retval 0 succes.
    \retval 1 failure.
*/
int
scas_log_time(int prty, char * fmt, ...)
{
    va_list     pvar;
    char        buf[MAX_LOG_BUFFER_SIZE];
    char *      datetime = NULL;
    char *      tmpbuf = NULL;
    int         res;
    time_t      mclock;
    struct tm * tmp = NULL;


    va_start(pvar, fmt);
    res=vsnprintf(buf,MAX_LOG_BUFFER_SIZE,fmt,pvar);
    va_end(pvar);
    if ( (res >= MAX_LOG_BUFFER_SIZE) || (res < 0) )
    {
        fprintf(stderr,"scas_log_time(): log string too long (> %d)\n",
                MAX_LOG_BUFFER_SIZE);
    }

    if (extra_logstr == NULL)
    {
        time(&mclock);
        /* tmp = localtime(&mclock); */
        tmp = gmtime(&mclock);

        datetime = malloc(sizeof(char) * 20);

        res=snprintf(datetime, 20, "%04d-%02d-%02d.%02d:%02d:%02d",
               tmp->tm_year + 1900, tmp->tm_mon + 1, tmp->tm_mday,
               tmp->tm_hour, tmp->tm_min, tmp->tm_sec);
        if ( (res >= 20) || (res < 0) )
        {
            fprintf(stderr,"scas_log_time(): date string too long (> %d)\n",
                    20);
        }

        tmpbuf = (char *) malloc ((strlen(datetime) + strlen(buf) + strlen(" : ")) * sizeof(char) + 1);
        strcpy(tmpbuf, datetime);
        strcat(tmpbuf, " : ");
        strcat(tmpbuf, buf);
    }
    else
    {
        tmpbuf = (char *) malloc ((strlen(extra_logstr) + strlen(buf) + strlen(" : ")) * sizeof(char) + 1);
        strcpy(tmpbuf, extra_logstr);
        strcat(tmpbuf, " : ");
        strcat(tmpbuf, buf);
    }

    if (logging_usrlog)
    {
        if (scas_logfp == NULL)
        {
            fprintf(stderr,"scas_log_time() error: cannot open file descriptor\n");
            return 1;
        }
        fprintf(scas_logfp,"SCAS %d: %s",prty,tmpbuf);
        fflush(scas_logfp);
    }
    if (logging_syslog && prty)
    {
        syslog(prty, tmpbuf);
    }

    if (datetime != NULL) free(datetime);
    if (tmpbuf != NULL) free(tmpbuf);

    return 0;
}


/******************************************************************************
Function:       scas_log_set_time_indicator()
Description:    Resets the log indicator to the current time
                When set, the logstring will be prefixed by this string
Parameters:
Returns:        
                
******************************************************************************/
int scas_log_set_time_indicator (void)
{
    time_t      _time   = 0;
    struct tm * _time_s = NULL;
    char        _strf_fmt[MAX_TIME_STRING_SIZE];
    int         _strf_size = 0;
    
    _time      = time(NULL);
    _time_s    = localtime(&_time);
    /* Note: strftime() returns number of bytes WITHOUT \0 */
    _strf_size = strftime(_strf_fmt, MAX_TIME_STRING_SIZE, "%F.%T%z", _time_s);
    if (_strf_size==0)	{
	fprintf(stderr,"strftime() returned 0, not changing extra_logstr\n");
	return 1;
    }

    /* Add \0 byte */
    _strf_size++;

    /* Free previous set extra_logstr */
    free(extra_logstr);
    
    /* Add locally to globally scope string */
    extra_logstr = malloc (sizeof (char) * (_strf_size));
    strncpy (extra_logstr, _strf_fmt, _strf_size);

    return 0;
}



/******************************************************************************
Function:       scas_get_log_file_pointer()
Description:    Get the file pointer to the logfile
Parameters:
Returns:        FILE * pointer to the open file
******************************************************************************/
FILE * scas_get_log_file_pointer(void)
{
    return scas_logfp;
}


/******************************************************************************
Function:       scas_get_log_type()
Description:    Get the active type of logging facility (syslog and/or file logging)
Parameters:
Returns:        unsigned short
******************************************************************************/
unsigned short scas_get_log_type (void)
{
    if (logging_syslog && logging_usrlog) return DO_SYSLOG|DO_USRLOG;
    if (logging_usrlog) return DO_USRLOG;
    if (logging_syslog) return DO_SYSLOG;

    return 0;
}

/******************************************************************************
Function:       scas_get_log_path()
Description:    Get the path string to the logfile
Parameters:
Returns:        char *
******************************************************************************/
char * scas_get_log_path (void)
{
    return scas_log_path;
}


/******************************************************************************
CVS Information:
    $Source: /srv/home/dennisvd/svn/mw-security/scas/src/scas-server/logging/scas_log.c,v $
    $Date: 2010-04-29 10:59:40 $
    $Revision: 1.10 $
    $Author: okoeroo $
******************************************************************************/
