/**
 * Copyright (c) Members of the EGEE Collaboration. 2007-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2007-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl> 
 *
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/wait.h>

#include "scas_config.h"

#include "daemonize.h"

#include "common.h"
#include "pdp_xacml_authz_process.h"

#include "xacml_server.h"
#include "xacml_io_ssl.h"
#include "ssl-common.h"
#include "net_common.h"


#include "open_conf_file.h"


#include "scas_log.h"
#include "_scas_log.h"


#define FORKCHILDMAX 10


#define ENABLE  1
#define DISABLE 0
#define TRUE    1
#define FALSE   0


#define SCAS_ERR_CMDLINE     10
#define SCAS_ERR_LOG_FAIL    20
#define SCAS_ERR_CONFIG_FAIL 30
#define SCAS_ERR_UNPRIV_PORT 40
#define SCAS_ERR_SYSCALLS    50

#define SCAS_PROCESS_UNINIT  0
#define SCAS_PROCESS_PARENT  1001
#define SCAS_PROCESS_CHILD   1002

#define SCAS_SERVICE_REFRESH_TIMER 600  /* In seconds */

static int parentprocess = SCAS_PROCESS_UNINIT;
static int mychild       = 0;
static int forcecheckpointer = 0;

static int done = 0;
static int gsig = 0;

static int service_refresh_timer = 0;


/* prototype */

int process_xacml (int listenport, int secured);
void siginthandler(int sig);
static int killchild(int childpid);
static int checkpoint(void);
int reinitlogfile(void);
void usage (void);


int reinitlogfile()
{
    unsigned short scas_logfile_type = 0;
    FILE * scas_logfile_fp = NULL;
    int    scas_logfile_fd = 0;
    char * scas_logfile_path = NULL;


    /* Get file pointer of the logfile and store the file descriptor */ 
    /* Only do this action if logging to file, skip procedure when logging to SYSLOG */
    if (DO_USRLOG == scas_get_log_type ())
    {
        scas_logfile_fp   = scas_get_log_file_pointer();
        scas_logfile_fd   = fileno (scas_logfile_fp);
        scas_logfile_type = scas_get_log_type();
        scas_logfile_path = scas_get_log_path();
        
   
        scas_log_debug (5, "SCAS: Re-opening the log facility... current FD is = %d, path is %s\n", scas_logfile_fd, scas_logfile_path);
    
        /* scas_log_close(); */
    
        /* Re-open log facility */
        if (scas_log_open (scas_logfile_path, NULL, scas_logfile_type))
        {
            fprintf (stderr, "SCAS logging facility failed to start up! Tried using the log file: %s\n", scas_logfile_path);
            return SCAS_ERR_LOG_FAIL;
        }
    
        /* close old descriptor */
        close (scas_logfile_fd);
    }

    return 0;
}

/*
 * Signal handler
 * This signal handler will stop the service gracefully
 */
void siginthandler(int sig)
{
    switch (sig)
    {
        case SIGTERM :
            /* Stop the service */

            /* Set time indicator to now! */
            scas_log_set_time_indicator();

            if (SCAS_PROCESS_PARENT == parentprocess)
            {
                scas_log (1, "SCAS: caught SIGTERM\n");
		/* Kill child and wait, ignore return value here */
		killchild(mychild);
            }
            gsig = sig;
            done = 1;
            break;
        case SIGHUP :
            /* Set time indicator to now! */
            scas_log_set_time_indicator();

            scas_log_debug (3, "SCAS: caught SIGHUP\n");
            /* Only the parent needs to reinit the log and re-issue a child process */
            /* Note: I'm againt child labor, but essentially this is what I've programmed here... */
            if (SCAS_PROCESS_PARENT == parentprocess)
            {
                forcecheckpointer = 1;
            }
            break;
        default :
            /* Stop the service */

            /* Set time indicator to now! */
            scas_log_set_time_indicator();

            scas_log (1, "SCAS: caught signal: %d\n", sig);
            if (SCAS_PROCESS_PARENT == parentprocess)
            {
		/* Kill child and wait, ignore return value here */
		killchild(mychild);
            }
            gsig = sig;
            done = 1;
    }
}

static int killchild(int childpid)	{
    int rc,save_errno;
    int status = 0;

    rc = kill(childpid, SIGTERM );
    save_errno=errno;
    if (rc<0)	{
	switch (rc)	{
	    case ESRCH: /* No process or process group with pid */
		scas_log(3, "Child process no longer exists\n");
		/* Still wait for the child, to prevent zombies */
		rc=waitpid(-1,&status,WNOHANG);
		break;
	    case EPERM: /* permission denied */
		scas_log(1, "Cannot send SIGTERM to childpid: %s\n",
			strerror(save_errno));
		return 1;
	    default:
		scas_log(1, "Killing child failed:%s\n",
			strerror(save_errno));
		return 1;
	}
    } else
	rc=wait(&status);
    if (rc<0)   {
	switch (rc)	{
	    case ECHILD:
		scas_log(3, "No terminated child is waiting\n");
		break;
	    case EINTR:
		scas_log(1, "wait got interrupted: %s\n",
			strerror(save_errno));
		return 1;
	    case EINVAL:
	    default:
		scas_log(1, "Waiting for child failed:%s\n",
			strerror(save_errno));
		return 1;
	}
    }

    return 0;
}

static int checkpoint()
{
    const int realtime_timeout = service_refresh_timer; /* seconds */
    time_t start;

    while (!done)   {
        start = 0;
	mychild  = fork(); /* Use global mychild pid var, to make sure we'll
			      always have a consistent idea about the PID */

        if (mychild == 0)
        {
            /* I'm the child process */
            /* I will (re)start working */
            parentprocess = SCAS_PROCESS_CHILD;

            /* Ignore the SIGHUP call explicitly from a child */
            /* A child process will inherit the registered signal handlers, 
             * which is unwanted behavior */
            signal (SIGHUP, SIG_IGN);
            
            return 0;
        }
        else if (mychild < 0)
        {
            /* something went wrong... need to bail out */

            /* Set time indicator to now! */
            scas_log_set_time_indicator();

            fprintf (stderr, "SCAS: Unable to gracefully restart due to low resources\n");
            exit (1);
        }
        else
        {
            /* I'm the parent */
            /* I'll wait for a signal, like SIGUSR1, to restart a new child as a replacement */
            /* wait (&status); */

            parentprocess = SCAS_PROCESS_PARENT;

            /* Timer */
            start = time(NULL);
	    scas_log(5,"Waiting %d sec before child process refresh\n",
		    realtime_timeout, forcecheckpointer);
            while ( ((time(NULL) - start) < realtime_timeout) &&
                    (!forcecheckpointer) && (!done) )
            {
                /* usleep (1000); */
                if (sleep (10)!=0)
		    scas_log(5,"sleep got interrupted\n");
            }

	    if (!done)	{
		/* Set time indicator to now! */
		scas_log_set_time_indicator();

		reinitlogfile();

		/* time's up, kill child */
		if (killchild(mychild)!=0)
		    return 1;
            
		forcecheckpointer = 0;
	    }
        }
    }

    return 0;
}


void usage (void)
{
    printf ("\n");
    printf (PACKAGE_STRING);
    printf ("\n\n");
    printf ("   Command line options:\n");
    printf ("   scas [-h|--h] [--conf <configuration file>] [--port <port number>] [--log <path to logfile>] [--cert <public certificate>] [--key <private key file>] [--capath <CA directory>] [--refresh-time <child process refresh interval in seconds>] [--no-daemon]\n");
    printf ("\n");
    printf ("                           -h|--h  -  The usage output you are looking at right now.\n");
    printf ("      --conf <configuration file>  -  Configuration file for the SCAS service.\n");
    printf ("                                      It is advised to primarily use the scas.conf for the configuration.\n");
    printf ("                                      See manpage for more about the the scas.conf options\n");
    printf ("             --port <port number>  -  TCP/IP port number to listen on.\n");
    printf ("          --log <path to logfile>  -  Sets the path to the log file that will be used.\n");
    printf ("      --cert <public certificate>  -  Sets the public certificate file for the SCAS service.\n");
    printf ("         --key <private key file>  -  Sets the private key file associated by the public.\n");
    printf ("                                      certificate file.\n");
    printf ("          --capath <CA directory>  -  Sets the Certificate Authority (CA) directory to be\n");
    printf ("                                      used for the CA public certificates and CRL files.\n");
    printf ("        --refresh-timer <seconds>  -  Sets the child process refresh interval in seconds.\n");
    printf ("                                      The parent process will wait this amount of seconds\n");
    printf ("                                      before refreshing the child process for a new one to\n");
    printf ("                                      keep memory leakage to a minimum.\n");
    printf ("                      --no-daemon  -  Keeps the service from daemonizing and refreshing. This\n");
    printf ("                                      option is very useful combined with debugging in GDB\n");
    printf ("\n");
    printf ("   Note: The command line options override the settings of the configuration file.\n");
    printf ("\n");
}



/* MAIN */
int main(int argc, char *argv[])
{
    int rc = 0;
    int i = 0;
    int portnumber = 0;
    int disableDaemonize = 0;
    char * log_filepath  = NULL;
    char * conf_filepath = NULL;

    config_name_values cnv = NULL;
    int                n_config_name_values = 0;


    certFilePub  = NULL;
    certFilePriv = NULL;
    certCAPath   = NULL;



    /* Commandline opts */
    for (i = 1; i < argc; i++)
    {
        if ((strcasecmp (argv[i], "--conf") == 0) && (i + 1 < argc))
        {
            conf_filepath = argv[++i]; 
        }
        else if (strcasecmp (argv[i], "--no-daemon") == 0)
        {
            disableDaemonize = 1;
        }
        else if ((strcasecmp (argv[i], "--port") == 0) && (i + 1 < argc))
        {
            portnumber = atoi (argv[++i]); 
        }
        else if ((strcasecmp (argv[i], "--log") == 0) && (i + 1 < argc))
        {
            log_filepath = argv[++i]; 
        }
        else if ((strcasecmp (argv[i], "--key") == 0) && (i + 1 < argc))
        {
            certFilePriv = argv[++i]; 
        }
        else if ((strcasecmp (argv[i], "--cert") == 0) && (i + 1 < argc))
        {
            certFilePub = argv[++i]; 
        }
        else if ((strcasecmp (argv[i], "--capath") == 0) && (i + 1 < argc))
        {
            certCAPath  = argv[++i]; 
        }
        else if ((strcasecmp (argv[i], "--refresh-timer") == 0) && (i + 1 < argc))
        {
            char * refresh_timer_tmp = argv[++i];

            service_refresh_timer = strtol(refresh_timer_tmp, NULL, 10);
            if (service_refresh_timer == 0)
            {
                fprintf (stderr, "SCAS: Couldn't parse --refresh-timer value. Doesn't seem to be a number: %s\n", refresh_timer_tmp);
                return SCAS_ERR_CMDLINE;
            }
            else if (service_refresh_timer < 1)
            {
                 fprintf (stderr, "SCAS: --refresh-timer value too low: %d\n", service_refresh_timer);
                 return SCAS_ERR_CMDLINE;
            }
        }
        else if ( (strcasecmp (argv[i], "--h") == 0) || (strcasecmp (argv[i], "-h") == 0) )
        {
            usage();
            return 0;
        }
        else
        {
            /* Unsupported option provided */
            fprintf (stderr, "SCAS: Unsupported commandline option found: %s\n", argv[i]);
            return SCAS_ERR_CMDLINE;
        }
    }


    signal(SIGINT, siginthandler);
    signal(SIGTERM, siginthandler);
    signal(SIGHUP, siginthandler);


    /* Open and read configuration file, if a configuration file is specified */
    if (conf_filepath)
    {
        if (open_scas_config_file(conf_filepath, &cnv, &n_config_name_values))
        {
            /* scas_log (1, "Couldn't open the configuration file %s\n", conf_filepath); */
            fprintf (stderr, "SCAS: Couldn't open the configuration file %s\n", conf_filepath);
            return SCAS_ERR_CONFIG_FAIL;
        }
    }
    else
    {
        scas_log (5, "No configuration file specified, only using commandline options\n");
    }


    /* Test routines by printing the structures */
    print_config_name_value (cnv, n_config_name_values);


    /* setup of SCAS, LCAS and LCMAPS environment variables */
    if (setup_scas_config_name_value (cnv, n_config_name_values))
    {
        fprintf (stderr, "SCAS: Couldn't setup SCAS, LCAS and LCMAPS configuration file options into the environment of this process.\n");
        /* scas_log (1,           "Couldn't setup SCAS, LCAS and LCMAPS configuration file options into the environment of this process.\n"); */
        return SCAS_ERR_CONFIG_FAIL;
    }

    /* Setting values that are not well expressed in environment variables here. 
       The origin of the value is the configuration file */
    /* If there is no commandline parameter set for the log file destination, then we'll look in the config file */
    if (!log_filepath)
    {
        if (!(log_filepath = get_scas_config_value_by_name (cnv, n_config_name_values, "scas_log_file")))
        {
            fprintf (stderr, "SCAS: Couldn't find the setting for the log file in either the configuration file or the commandline parameters.\n");
            return SCAS_ERR_CONFIG_FAIL;
        }
    }

    
    /* Open Log file facility */
    if (scas_log_open (log_filepath, NULL, DO_USRLOG))
    {
        fprintf (stderr, "SCAS logging facility failed to start up! Tryed using the log file: %s\n", log_filepath);
        return SCAS_ERR_LOG_FAIL;
    }

    /* Set time indicator to now! */
    scas_log_set_time_indicator();

    /* SCAS service child process refresh rate selection */
    if (service_refresh_timer == 0)
    {
        char * tmp_refresh_rate = NULL;

        if (!(tmp_refresh_rate = get_scas_config_value_by_name (cnv, n_config_name_values, "scas_refresh_rate")))
        {
            service_refresh_timer = SCAS_SERVICE_REFRESH_TIMER;
        }
        else
        {
            service_refresh_timer = strtol(tmp_refresh_rate, NULL, 10);
            if (service_refresh_timer == 0)
            {
                fprintf (stderr, "SCAS: Couldn't parse scas_refresh_rate value in the configuration file. Doesn't seem to be a number: %s\n", tmp_refresh_rate);
                return SCAS_ERR_CONFIG_FAIL;
            }
        }
        scas_log (5, "Using a child process refresh interval of %d seconds\n", service_refresh_timer);
    }

    /* Daemonization */
    /* Disable the daemonization (for debugging purposes) */
    /* disableDaemonize = 1; */

/*
 *   if (!disableDaemonize)
 *   {
 *
 *       char * tmp_choice = NULL;
 *
 *       if ((tmp_choice = get_scas_config_value_by_name (cnv, n_config_name_values, "scas_no_daemonize")))
 *       {
 *           if (strncasecmp ("yes", tmp_choice, strlen("yes")) == 0)
 *           {
 *               disableDaemonize = 1;
 *               scas_log_debug (5, "Using \'no daemonizing\' mode\n");
 *           }
 *       }
 *   } 
 */


    /* Portnumber selection */
    if (portnumber == 0)
    {
        char * tmp_port_num = NULL;

        if (!(tmp_port_num = get_scas_config_value_by_name (cnv, n_config_name_values, "scas_port")))
        {
            fprintf (stderr, "SCAS: Couldn't find the setting for the listening portnumber in the configuration file or the commandline parameters.\n");
            return SCAS_ERR_CONFIG_FAIL;
        }
        else
        {
            portnumber = atoi (tmp_port_num);
            scas_log_debug (5, "Using portnumber %d\n", portnumber);
        }
        free (tmp_port_num);
    }

    /* CA path */
    if (!certCAPath)
    {
        if (!(certCAPath = get_scas_config_value_by_name (cnv, n_config_name_values, "scas_capath")))
        {
            fprintf (stderr, "SCAS: Couldn't find the setting for the CA Path which holds the CA certificates in the configuration file or the commandline parameters.\n");
            return SCAS_ERR_CONFIG_FAIL;
        }
        else
            scas_log_debug (5, "Using CA Path: %s\n", certCAPath); 
    } 

    /* Host / service certificate */
    if (!certFilePub)
    {
        if (!(certFilePub = get_scas_config_value_by_name (cnv, n_config_name_values, "scas_hostcert")))
        {
            fprintf (stderr, "SCAS: Couldn't find the setting for the Host / Service certificate file in the configuration file or the commandline parameters.\n");
            return SCAS_ERR_CONFIG_FAIL;
        }
        else
            scas_log_debug (5, "Using Public certificate: %s\n", certFilePub); 
    } 

    /* Host / service private key file */
    if (!certFilePriv)
    {
        if (!(certFilePriv = get_scas_config_value_by_name (cnv, n_config_name_values, "scas_hostkey")))
        {
            fprintf (stderr, "SCAS: Couldn't find the setting for the Host / Service private key file in the configuration file or the commandline parameters.\n");
            return SCAS_ERR_CONFIG_FAIL;
        }
        else
            scas_log_debug (5, "Using Private key: %s\n", certFilePriv); 
    } 


    /* Test portnummer appliance, only use first 1024 ports if your root */
    if ((geteuid() != 0) && (portnumber < 1025))
    {
        fprintf (stderr, "SCAS: Without being Root unable to use a port below 1024. Requested portnumber to use: %d\n", portnumber);
        return SCAS_ERR_UNPRIV_PORT;
    }


    /* Properly daemonize the service (if not disabled) */
    /* Any return value other then zero is not good */
    /* Check point */
    if (disableDaemonize)
    {
        fprintf (stderr, "SCAS: Disable daemonization\n");
    }
    else
    {
        /* if (!daemonize() && !checkpoint()) */
        if (daemonize())
        {
            fprintf (stderr, "SCAS: Couldn't perform the daemonization.\n");
            return SCAS_ERR_SYSCALLS;
        }
        /* Start checkpointing from here, and revive the new children here */   
        if (checkpoint())
        {
            fprintf (stderr, "SCAS: Couldn't perform the checkpointing.\n");
            return SCAS_ERR_SYSCALLS;
        }
    }

    /* Only start processing XACML in child process or non-daemonize, both cases
     * mychild==0 */
    if (mychild==0) {
	/* Start an application wide mutex, used to channel the LCMAPS execution
	 * because its not thread safe yet */
	pthread_mutex_init(&mutex, NULL);

	scas_log_debug(4, "Starting process_xacml\n");
	rc = process_xacml (portnumber, ENABLE);

	pthread_mutex_destroy (&mutex);
    }
    
    if (mychild!=0 || disableDaemonize)
	scas_log(3, "SCAS finished\n");

    /* Successfully setup environment, going to remove the allocated memory */
    free_config_name_value (&cnv, n_config_name_values); 

    scas_log_close();
    return rc;
}



/********************************************************************
function:       process_xacml
description:
                This function is the start of the service. It initializes the XACML library
                and will put the library in listen mode. In this case that I/O layer is hardwired to be secured
                via my I/O handler function implementation, that brings you the SSL layer. And in effect, HTTPS.

paramters:
      int listenport    :    The service will listen on this TCP port number
      int secured       :    This is the switch between secured (via I/O handlers) or plain text HTTP. 
********************************************************************/ 
int process_xacml (int listenport, int secured)
{
    int rc = 0;
    char * logstr = "process_xacml";
#ifdef HAVE_XACML_SERVER_SET_FD
    int master_socket = -1;
#else
    unsigned short port = listenport;
#endif /* HAVE_XACML_SERVER_SET_FD */
    xacml_server_t server;
    /* xacml_response_t response; */



    /******* SSL Init *********/
    if (init_OpenSSL())
    {
        scas_log (1, "%s: Error: Failed to initialize the OpenSSL library\n", logstr);
        return 1;
    }
    else
    {
        scas_log_debug (3, "%s: OpenSSL library initialized\n", logstr);
    }

    seed_prng();


    ssl_ctx = setup_server_ctx(NULL, certCAPath, certFilePub, certFilePriv);
    if (!ssl_ctx)
    {
        scas_log (1,     "%s: Error: unable to create an SSL_CTX context with setup_server_ctx()\n", logstr);
        fprintf (stderr, "%s: Error: unable to create an SSL_CTX context with setup_server_ctx()\n", logstr);
        rc = 0;
        goto process_xacml_cleanup;
    }
    /**************************/


    xacml_init();
    xacml_server_init(&server, xacml_authorize, NULL);

    /* Set server object properties to listen to give en port */


#ifdef HAVE_XACML_SERVER_SET_FD
    /* Setting up master socket */
    if (setup_master_socket (&master_socket, listenport))
    {
        scas_log (1,     "SCAS setup failure: unable to setup listening socket on port %d\n", listenport);
        fprintf (stderr, "SCAS setup failure: unable to setup listening socket on port %d\n", listenport);

        /* not ok */
        rc = 1;
        goto process_xacml_cleanup;
    }


    /* Setting the listen()-ing master socket. */
    if (xacml_server_set_fd (server, master_socket) != XACML_RESULT_SUCCESS)
    {
        scas_log (1,     "SCAS setup failure: Unable to register the listening socket into the SAML2-XACML2 library.\n");
        fprintf (stderr, "SCAS setup failure: Unable to register the listening socket into the SAML2-XACML2 library.\n");

        /* not ok */
        rc = 1;
        goto process_xacml_cleanup;
    }
        
#else
    xacml_server_set_port(server, listenport);

    xacml_server_get_port(server, &port);
    scas_log_debug (2, "SCAS ready... listening on port %hu.\n", port);
    printf         (   "SCAS ready... listening on port %hu.\n", port);
#endif /* HAVE_XACML_SERVER_SET_FD */


    /* Set the I/O handler object */
    xacml_server_set_io_descriptor(server, &xacml_io_ssl_descriptor);
     
    xacml_server_start(server);

    /* Set time indicator to now! */
    scas_log_set_time_indicator();

    while (!done)
    {
        pause();
        printf ("SCAS service stop.\n");
    }

    fprintf(stderr, "SCAS: Shutting down... signal was: %d\n", gsig);
    scas_log (3,    "Shutting down... signal was: %d\n", gsig);

    xacml_server_destroy(server);
    fflush(stdout);
    fflush(stderr);

    /* Perfect ending */
    rc = 0;

process_xacml_cleanup:
    SSL_CTX_free (ssl_ctx);

    return rc;
}
